# known
__doc__=r"""
**known** is a collection of reusable python code. Use pip to install **known**

.. code-block:: console

   $ pip install known
   
The package is frequently updated by adding new functionality, make sure to have the latest version. To check version use

.. code-block:: console

   >>> import known
   >>> known.__version___

:py:mod:`known/__init__.py`
"""
__version__ = '0.0.7'
#print(f'known.{__version__}')


# from .basic import *
# from .imgu import *
# from .mailer import *
# from .store import *

