from setuptools import setup

setup(
    name =                      'known',
    version =                   '0.0.7',
    url =                       'https://github.com/Nelson-iitp/known',
    author =                    'Nelson.S',
    author_email =              'mail.nelsonsharma@gmail.com',
    description =               'Module :: known',
    packages =                  ['known'],
    classifiers=                ['License :: OSI Approved :: MIT License'],
    package_dir =               { '' : 'module'},
    install_requires =          [],
    include_package_data =      True
)   